//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//% CLIP2TXT.C - Copy the contents of the Windows Clipboard to a named file %
//% Written by James Greene Informatics Consulting, Lrrach, Germany        %
//%-------------------------------------------------------------------------%
//% Last Modified:                                                          %
//%     21.01.1999 - Release of version 1.0                                 %
//%     05.01.2015 - Updated e-mail and contact information, compiled with  %
//%                  gcc 4,8.1.                                             %
//%     09.11.2016 - Updated to support various text formats.               %
//%-------------------------------------------------------------------------%
//% CLIP2TXT Version 2.0 (09. November 2016)                                %
//% Copyright (c) 1999, 2015, 2016 by James Greene. All rights reserved.    %
//% CLIP2TXT comes with NO WARRANTY to the extent permitted by law.         %
//% You may redistribute copies of CLIP2TXT under the terms of the          %
//% GNU General Public License. For more information about these            %
//% matters, see the file named COPYING, or contact the author at           %
//% <james.greene@outlook.com>                                              %
//%                                                                         %
//% Please report any bugs to <james.greene@outlook.com>                    %
//%                                                                         %
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <conio.h>
#include <errno.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#define VER_NUMBER  "2.02"
#define VER_STRING  "Version " VER_NUMBER
#define VER_DATE    "06. December 2016"

static char *strUsage[] = {
	//	"....-....1....-....2....-....3....-....4....-....5....-....6....-....7....-....8",
	    "Options:  ( /option or --option are also accepted )",
	    "  -D[isplay]   = Display the clipboard contents on the screen.",
	    "  -H[elp]      = Display this usage information.",
	    "  -?           = Same as -help.",
	    "  -N[ologo]    = Don't display the Copyright banner.",
	    "  -Q[uiet]     = Same as -nologo.",
	    "  -S[ilent]    = Same as -nologo.",
		"  -F[ormat]=[xyz] where [xyz] = ASC or DOS, ANS or TXT, HTM, UNI, RTF or C2T.",
		"                  for ASCII, ANSI, HTML, UNICODE, Rich Text Format or CLIP2Txt.",
	    "  -T[ranslate] = (depreciated) same as -FORMAT:DOS.",
		"",
	    "  The following options are all interpreted the same:",
	    "     -d, /d, -DISP, /Disp, --DISPLAY, -dIsP, ... you get the idea!\n",
	    "  If no filename is specified, the Windows Clipboard contents will be displayed",
	    "  on the screen, so CLIP2TXT > FOO.TXT is equivalent to CLIP2TXT -F=DOS FOO.TXT",
		""
};

#define	numUsage (sizeof(strUsage) / sizeof(char *))

static char *strVer[] = {
//	"....-....1....-....2....-....3....-....4....-....5....-....6....-....7....-....8",
	"Copyright \xb8 1999, 2015, 2016 James Greene. All rights reserved.",
	"CLIP2TXT comes with NO WARRANTY to the extent permitted by law. You may",
	"redistribute copies of CLIP2TXT under the terms of the GNU General Public",
	"License. For more information about these matters, see the file named",
	"COPYING.txt, or contact the author at <james.greene@outlook.com>.",
	"",
	"Please report any bugs to <james.greene@outlook.com>.",
	""
};

#define	numVer (sizeof(strVer) / sizeof(char *))

#define CF_C2TFORM	0xF045		// proprietary CLIP2Txt Clipboard dump file

#define idxHTM 	0
#define idxRTF 	1
#define idxTXT  2
#define idxDOS  4
#define idxUNI	6
#define idxC2T	7

static struct cfOPT {
	unsigned cfVal;
	unsigned cfInt;
	char *cfName;
	char *cfDisp;
} cfValues[] = {
	{0, idxHTM, "HTM", "HTML Format"},			// HTM = 0
	{0, idxRTF, "RTF", "Rich Text Format"},			// RTF = 1
	{CF_TEXT, idxTXT, "ANS", "Text Format"},
	{CF_TEXT, idxTXT, "TXT", "Text Format"},
	{CF_OEMTEXT, idxDOS, "ASC", "OEM Text Format"},
	{CF_OEMTEXT, idxDOS, "DOS", "OEM Text Format"},
	{CF_UNICODETEXT, idxUNI, "UNI", "Unicode Text Format"},
	{CF_C2TFORM, idxC2T, "C2T", "CLIP2Txt Format"},
};

#define numCFOPT	(sizeof(cfValues) / sizeof(struct cfOPT))

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

typedef struct tagOPTIONS {
	unsigned fQuiet: 1;
	unsigned fDisplay: 1;
	unsigned fHelp: 1;
	unsigned fVersion: 1;
    unsigned uFormat;
    unsigned cfHTML;
    unsigned cfRTF;
    char szName[MAX_PATH];
} OPTIONS;

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static int my_strnicmp(const char *pStr1, const char *pStr2, size_t Count) {
    char c1, c2;
    int v;

    if (Count == 0)
        return 0;

    do {
        c1 = *pStr1++;
        c2 = *pStr2++;
        /* the casts are necessary when pStr1 is shorter & char is signed */
        v = (unsigned) tolower(c1) - (unsigned) tolower(c2);
    } while ((v == 0) && (c1 != '\0') && (--Count > 0));

    return v;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static  void    AnyKey(void) {
    register int ri;
    char szMsg[32];

    strcpy(szMsg, "(press any key to continue)");
    
    fprintf(stderr, szMsg);
    fflush(stderr);
    
    if (!getch())
        getch();
        
    for (ri=strlen(szMsg) ; ri > 0; ri--) 
        fprintf(stderr, "\b \b");
    fflush(stderr);
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static void Version(void) {
	int ri;
    fprintf(stderr, "CLIP2TXT Version %s (%s at %.5s)\n", VER_NUMBER, VER_DATE, __TIME__);
    for (ri = 0; ri < numVer; ri++) {
    	fputs(strVer[ri], stderr);
    	fputs("\n", stderr);
    }
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static  void    Usage(char *exe) {
	int ri;
    fprintf(stderr, "Usage: %s [options] filename [options]\n", exe);
    for (ri = 0; ri < numUsage; ri++) {
    	fputs(strUsage[ri], stderr);
    	fputs("\n", stderr);
    }

    AnyKey();
    
				//	"....-....1....-....2....-....3....-....4....-....5....-....6....-....7....-....8",
    fprintf(stderr, "Examples:\n");
    fprintf(stderr, "  %s FOO.TXT\n", exe);
    fprintf(stderr, "    Copies the Windows Clipboard contents to the file FOO.TXT in the current\n");
    fprintf(stderr, "    directory.\n");
    fprintf(stderr, "  %s -f:txt \"C:\\My Documents\\Long File Name.txt\"\n", exe);
    fprintf(stderr, "    Copies the Clipboard contents to the specified file. Filenames containing\n");
    fprintf(stderr, "    spaces must be \"enclosed in double quotes.\"\n");
    fprintf(stderr, "  %s -form:rtf document.rtf\n", exe);
    fprintf(stderr, "    Copies the RTF version of the Windows Clipboard to the file 'document.rtf'.\n\n");

    Version();
}
    
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static  BOOL ParseCommandLine(char *argv[], OPTIONS *opt) {
    char *pgm, *arg;
    register int ri;

    memset(opt, 0, sizeof(OPTIONS));
    opt->uFormat = CF_TEXT;

    // query Clipboard formats
	opt->cfRTF = RegisterClipboardFormat("Rich Text Format");
	opt->cfHTML = RegisterClipboardFormat("HTML Format");

	// update the cfValues array
	cfValues[idxHTM].cfVal = opt->cfHTML;
	cfValues[idxRTF].cfVal = opt->cfRTF;

    pgm = *argv;
    
    while ((arg = *++argv) != NULL) {
    	if (*arg == '\'')	// workaround for my debugger
    		++arg;

        if ((*arg == '/') || (*arg == '-')) {   // flag
            ri = strlen(++arg);
            if (*arg == '-')
                ri = strlen(++arg);

            if (toupper(*arg) == 'F') {
            	char *brk = strpbrk(arg, ":=");
            	if (brk != 0) {
            		int fri = brk - arg;
            		++brk;

            		if (my_strnicmp(arg, "FORMAT", fri) == 0) {
            			int lv;
            			int savFormat = opt->uFormat;
            			opt->uFormat = 999999;
            			for (lv = 0; (lv < numCFOPT) ; lv++) {
            				if (my_strnicmp(brk, cfValues[lv].cfName, 3) == 0) {
            					opt->uFormat = cfValues[lv].cfVal;
            					break;
            				}
            			}

            			if (opt->uFormat == 999999) {
            				fprintf(stderr, "CLIP2TXT: Unrecognized format '%s'\n", brk);
            				opt->uFormat = savFormat;
            			}
            		}
            	} else {
                    fprintf(stderr, "CLIP2TXT: Option '%s' not recognized - ignored\n", --arg);
            	}
            }
            else if (my_strnicmp(arg, "TRANSLATE", ri) == 0) {
            	if (opt->uFormat == CF_TEXT)
            		opt->uFormat = CF_OEMTEXT;
            }
            else if (my_strnicmp(arg, "DISPLAY", ri) == 0)
                opt->fDisplay = TRUE;
            else if ((my_strnicmp(arg, "NOLOGO", ri) == 0) ||
                      (my_strnicmp(arg, "QUIET", ri) == 0) ||
                      (my_strnicmp(arg, "SILENT", ri) == 0))
                opt->fQuiet = TRUE;
            else if ((my_strnicmp(arg, "HELP", ri) == 0) ||
                      (my_strnicmp(arg, "?", ri) == 0))
                opt->fHelp = TRUE;
            else if ((my_strnicmp(arg, "VERSION", ri) == 0))
                opt->fVersion = TRUE;
            else
                fprintf(stderr, "CLIP2TXT: Option '%s' not recognized - ignored\n", --arg);
        } else {
            if (!*(opt->szName))
                strcpy(opt->szName, arg);
            else
                fprintf(stderr, "CLIP2TXT: Too many parameters: '%s' ignored\n", arg);
        }
    }

    if (!opt->fQuiet) {
        fprintf(stderr, "\n%-65s %14s\n", "CLIP2TXT: Copy the Windows Clipboard contents to a File", VER_STRING);
        fprintf(stderr, "Copyright \xb8 1999, 2015, 2016 by James Greene. All rights reserved.\nASP Member #06128.\n");
        fprintf(stderr, "%38s * %-38s\n\n", "http://informatics-consulting.de", "james.greene@outlook.com");
    }

    // if no filename specified, then just display the text to the screen
    if (!*(opt->szName)) {
    	if(opt->uFormat == CF_TEXT)
    		opt->uFormat = CF_OEMTEXT;
        opt->fDisplay = TRUE;
    }

    if (opt->fVersion) 
        Version();
        
    if (opt->fHelp) {
        Usage(pgm);
        return FALSE;
    }

    return TRUE;
}
        
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

char	*cbFormName(OPTIONS *opt) {
	UINT lv;

	for (lv = 0; (lv < numCFOPT) ; lv++)
		if (opt->uFormat == cfValues[lv].cfVal)
			return cfValues[lv].cfDisp;

	return "*Unknown* Format";
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#define IDS_LOCKFAIL	0x8001
#define IDS_ALLOCFAIL	0x8002

HGLOBAL GetCBData(OPTIONS *opt, PDWORD pdwSize) {
	HGLOBAL hClp, hData = INVALID_HANDLE_VALUE;
	DWORD dwSize;
	UINT uSize;
	LPBYTE lpC, lpN;
	int nMsg = 0;

	if ((hClp = GetClipboardData(opt->uFormat)) != 0) {
		uSize = GlobalFlags(hClp);
		if (uSize & GMEM_DISCARDED)		// if it was discarded, lock it so that it is called back into memory
			lpC = GlobalLock(hClp);
		dwSize = GlobalSize(hClp);		// now the size should be correct
		if (uSize & GMEM_DISCARDED)
			GlobalUnlock(hClp);

		if (pdwSize)
			*pdwSize = dwSize;

		if (!dwSize)
			return hData;

		uSize = 2 + dwSize;

		if ((lpC = GlobalLock(hClp)) != 0) {
			if ((hData = GlobalAlloc(GMEM_MOVEABLE, (SIZE_T)uSize)) != 0) {
				if ((lpN = GlobalLock(hData)) != 0) {
					CopyMemory(lpN, lpC, dwSize);
					*(PTSTR)(lpN+dwSize) = 0;
					GlobalUnlock(hData);
				} else {
					nMsg = IDS_LOCKFAIL;
					GlobalFree(hData);
					hData = INVALID_HANDLE_VALUE;
				}
			} else {
				hData = INVALID_HANDLE_VALUE;
				nMsg = IDS_ALLOCFAIL;
			}

			GlobalUnlock(lpC);
		} else
			nMsg = IDS_LOCKFAIL;
	} else {
		fprintf(stderr, "CLIP2TXT: The Windows Clipboard does not contain %s data!\n", cbFormName(opt));
		nMsg = 0;
	}

	if (nMsg == IDS_ALLOCFAIL)
        fprintf(stderr, "CLIP2TXT: Unable to allocate memory for Windows Clipboard contents!\n");
	else if (nMsg == IDS_LOCKFAIL)
        fprintf(stderr, "CLIP2TXT: Unable to lock the Windows Clipboard storage!\n");

	return hData;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// assumptions:
//	- the Clipboard contained valid HTML text (including the header information)
// 	- fp is a valid file handle
// 	- the global memory (pointed to by lp) has been opened and locked
//	- dwSize is the size of the block

static	UINT	ClipHTML(FILE *fp, LPSTR lp, DWORD dwSize) {
	UINT uStart=0, uEnd=dwSize;
	UINT uLen;
	LPSTR lpLocal = lp;
	BYTE by;
	PSTR sTemp, sKey, sVal, sStop;

	// parse the header information
	while (*lpLocal && (*lpLocal != '<')) {
		LPSTR lpEnd = lpLocal + 1;
		while (*lpEnd && (*lpEnd != '\n'))
			++lpEnd;
		by = *lpEnd;
		*lpEnd = 0;
		sTemp = (PSTR)strdup(lpLocal);
		*lpEnd = by;
		lpLocal = lpEnd + 1;

		sKey = strtok(sTemp, ":");
		sVal = strtok(NULL, "\r");

		if (my_strnicmp(sKey, "StartHTML", 9) == 0)
			uStart = strtol(sVal, &sStop, 10);
		else if (my_strnicmp(sKey, "EndHTML", 7) == 0)
			uEnd = strtol(sVal, &sStop, 10);

		free(sTemp);
	}

	uLen = uEnd - uStart;
    fwrite(lp+uStart, sizeof(BYTE), uLen, fp);

    return uLen;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// This function is called if the "C2T" dump format was specified. In this case,
// we'll write a binary file which contains all of the available (text) Clipboard
// formats, so that the whole Clipboard can be restored by TEXT2Clp.

#define C2TSIG	0xA6543243		// "C2T"
struct _c2t_seg	{
	DWORD dwSig;
	unsigned uFmt;
	DWORD dwStart;
	DWORD dwLen;
};

static	BOOL	ClipDump(OPTIONS *opt) {
	UINT uCurFmt = 0;
	FILE *fp;
	HANDLE hd;
	DWORD dwSize, dwPos = 0L;
	struct _c2t_seg c2tseg;
	LPSTR lp;
	BOOL rv = FALSE;

    if (!*opt->szName) {
        fprintf(stderr, "CLIP2TXT: You must specify a filename when using the -F:C2T option.\n");
        return FALSE;
    }

    if (!OpenClipboard(NULL)) {
        fprintf(stderr, "CLIP2TXT: Unable to open the Windows Clipboard!\n");
        return FALSE;
    }

	fp = fopen(opt->szName, "wb");
	if (!fp) {
		fprintf(stderr, "CLIP2TXT: Unable to open file %s: %s\n",
				opt->szName, strerror(errno));
		CloseClipboard();
		return FALSE;
	}

	if (opt->fDisplay)
		fprintf(stderr, "CLIP2TXT: Please note that -DISP is not supported when using -FORM:C2T\n");

    while ((uCurFmt = EnumClipboardFormats(uCurFmt)) != 0) {
    	// check if we can handle this format
    	int lv;

		for (lv = 0; (lv < numCFOPT) && (uCurFmt != cfValues[lv].cfVal); lv++) ;
		if ((lv < numCFOPT) && (uCurFmt == cfValues[lv].cfVal)) {
			// we have a format we can deal with
			opt->uFormat = uCurFmt;
		    hd = GetCBData(opt, &dwSize);
		    c2tseg.dwSig = C2TSIG;
		    c2tseg.uFmt = cfValues[lv].cfInt;
		    c2tseg.dwLen = dwSize;
			c2tseg.dwStart = dwPos + sizeof(c2tseg);

			lp = (LPSTR)GlobalLock(hd);
			if (lp) {
				if (fp) {
					dwPos += fwrite(&c2tseg, sizeof(c2tseg), 1, fp);
					dwPos += fwrite(lp, sizeof(BYTE), dwSize, fp);
				    if (!opt->fQuiet)
				    	fprintf(stderr, "CLIP2TXT: %lu Bytes (%s) written to '%s'\n", dwSize, cfValues[lv].cfDisp, opt->szName);
				}

				GlobalUnlock(hd);
				GlobalFree(hd);
				rv = TRUE;			// flag success
			} else     // this should NEVER happen!
				fprintf(stderr, "CLIP2TXT: Unable to lock the Clipboard memory!\n");
		}
    }

    fclose(fp);

    CloseClipboard();

    if (!opt->fQuiet)
         fprintf(stderr, "CLIP2TXT: %lu Bytes written to '%s'\n", dwPos, opt->szName);

	opt->uFormat = CF_C2TFORM;
    return rv;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static  BOOL    ClipCopy(OPTIONS *opt) {
    HGLOBAL hd;
    LPSTR lp;
    FILE *fp;
    BOOL rv = FALSE;
    DWORD dwSize;
    UINT uOut = 0;

    // special handler for the C2F dump file
    if (opt->uFormat == CF_C2TFORM)
    	return ClipDump(opt);

    if (!OpenClipboard(NULL)) {
        fprintf(stderr, "CLIP2TXT: Unable to open the Windows Clipboard!\n");
        return FALSE;
    }

    hd = GetCBData(opt, &dwSize);
    if (hd == INVALID_HANDLE_VALUE) {
        CloseClipboard();
        return FALSE;
    }

    if (*opt->szName) {
    	fp = fopen(opt->szName, "wb");
        if (!fp) {
            fprintf(stderr, "CLIP2TXT: Unable to open file %s: %s\n",
                    opt->szName, strerror(errno));
            GlobalFree(hd);
            CloseClipboard();
            return FALSE;
        }
    } else
        fp = NULL;

    lp = (LPSTR)GlobalLock(hd);
    if (lp) {
        if (fp) {
        	if (opt->uFormat == opt->cfHTML)
        		uOut = ClipHTML(fp, lp, dwSize);
			else
	            uOut = fwrite(lp, sizeof(BYTE), dwSize, fp);
        }

        if (opt->fDisplay)
            fwrite(lp, sizeof(BYTE), dwSize, stdout);
        else if (fp && !opt->fQuiet)
             fprintf(stderr, "CLIP2TXT: %u Bytes written to '%s'\n", uOut, opt->szName);

         GlobalUnlock(hd);
         rv = TRUE;			// flag success
    } else     // this should NEVER happen!
        fprintf(stderr, "CLIP2TXT: Unable to lock the Clipboard memory!\n");
        
    if (fp)
        fclose(fp);
    if (hd != INVALID_HANDLE_VALUE)
    	GlobalFree(hd);
    CloseClipboard();

    return rv;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
int main(int argc, char *argv[]) {
    OPTIONS opt;

    return ParseCommandLine(argv, &opt) && ClipCopy(&opt) ? 1 : 0;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

